module abstract_conjugate_gradients

  ! cg is a polymorphic conjugate gradients solver that works with vectors and
  ! gradients in the VECTOR_CLASS and GRADIENT_CLASS classes.

contains
  subroutine cg (x_initial,g_initial,x_final,g_final,n_iters,simulator,ipmin)

  ! Inputs: x_initial = initial guess at solution vector.
  !         g_initial = initial gradient (with respect to any inner product).
  !         n_iters   = number of iterations to perform. In this demonstration
  !                     code, this is the only stopping criterion!)
  !         simulator = a function taking a vector argument and an inner
  !                     product, and returning a graident with respect to
  !                     that inner product.

  ! Outputs: x_final = solution vector
  !          g_final = gradient at the solution point with respect to the
  !                    same inner product as g_initial. This inner product
  !                    may be different from that specified by the argument
  !                    ipmin.

  use abstract_vector
  use abstract_gradient
  use abstract_inner_product
  use abstract_simulator

  implicit none

  class(vector_class),    intent(in)    :: x_initial
  class(vector_class),    intent(inout) :: x_final
  class(gradient_class),  intent(in)    :: g_initial
  class(gradient_class),  intent(inout) :: g_final
  integer,                intent(in)    :: n_iters
  class(simulator_class), intent(in)    :: simulator
  class(inner_product_class)            :: ipmin

  integer :: j
  real :: alpha, beta, gnorm2, gnorm2_new, dnorm
  class(vector_class),        allocatable :: d, w
  class(gradient_class),      allocatable :: g, z
  class(inner_product_class), pointer     :: ip_save

  !--- change inner product
  ip_save => g_initial%inquire_inner_product()
  call g_initial%change_inner_product_to(ipmin)

  !--- allocate work space. NB: this will only work for simple types.
  !--- The correct way to do it is to define an allocate method in
  !--- vector_class.

  allocate(d,source=x_initial)
  allocate(w,source=x_initial)
  allocate(g,source=g_initial)
  allocate(z,source=g_initial)

  call g%assign(g_initial)              ! g=g_initial

  !--- initial descent direction
  call g%to_vector(d)                   
  call d%negate()                       ! d=-g

  gnorm2 = ipmin%dot_g_g(g,g)

  print *,'initial gradient norm squared=',gnorm2, &
        & ' (',ipmin%name,' inner product)'

  !--- conjugate gradients...
  do j=1,n_iters
    dnorm = sqrt(ipmin%dot_v_v(d,d))

    call w%assign(x_initial)            ! w = x_initial
    call w%saxpy((1/dnorm),d)           ! w = w + (1/dnorm)*d

    call simulator%eval(w,z,ipmin)      ! z = grad_J(w)

    call z%subtract(g_initial)          ! z = z - g_initial
    call z%multiply_by(dnorm)           ! z = z*dnorm = J"d

    call z%to_vector(w)
    alpha = gnorm2/ipmin%dot_v_v(d,w)   ! alpha = <g,g>/<d,J"d>

    call x_final%saxpy(alpha,d)         ! x_final = x_final + d*alph

    call g%saxpy(alpha,z)               ! g = g + alpha*J"d

    gnorm2_new = ipmin%dot_g_g(g,g)
    beta = gnorm2_new/gnorm2            ! beta = <g,g>/<g_old,g_old>
    gnorm2 = gnorm2_new

    print *,'iteration ',j,' gradient norm squared=',gnorm2, &
        & ' (',ipmin%name,' inner product)'

    call d%multiply_by(beta)
    call g%to_vector(w)
    call d%subtract(w)                   ! d = -g + d*beta
  enddo

  !--- revert to original inner product

  call g_final%assign(g)                 ! g_final = g

  call g_initial%change_inner_product_to(ip_save)
  call g_final%change_inner_product_to(ip_save)

  !--- deallocate work space
  deallocate(d)
  deallocate(g)
  deallocate(w)
  deallocate(z)

  end subroutine cg
end module abstract_conjugate_gradients
