module abstract_gradient
  use :: abstract_vector
  use :: abstract_inner_product
  implicit none

  private
  public :: gradient_class

! Mathematically, a gradient is a vector defined with respect to an
! inner product. Changing the inner product results in a corresponding
! change in the vector.

! GRADIENT_CLASS is an abstract type that defines gradients.

! GRADIENT_CLASS adds the following data:
!   my_inner_product : (private) A pointer to the inner product with respect
!                      to which the gradient is currently defined.

! GRADIENT_CLASS adds the following methods:
!   change_inner_product_to(this,newdot) : changes the inner product with
!                                          respect to which the gradient is
!                                          defined. The elements of this are
!                                          transformed appropriately.
!   change_inner_product_to(this,newdot) : sets the inner product for the
!                                          gradient. The elements are not
!                                          transformed.
!   inquire_inner_product(this)          : returns a pointer to the current
!                                          inner product.
!   to_vector(this)                      : convert a gradient to a vector

  type, abstract, extends(vector_class) :: gradient_class
    private
    class(inner_product_class), pointer :: my_inner_product => null()
  contains
    private
    procedure, public, non_overridable  :: change_inner_product_to 
    procedure, public, non_overridable  :: set_inner_product
    procedure, public, non_overridable  :: inquire_inner_product
    procedure(op_g_v), public, deferred :: to_vector
  end type gradient_class

  abstract interface
    subroutine op_g_v(this,v)
      import vector_class
      import gradient_class
      class(gradient_class), intent(in)    :: this
      class(vector_class),   intent(inout) :: v
    end subroutine
  end interface

contains

  subroutine change_inner_product_to (this,newdot)
    implicit none
    class(gradient_class)                :: this
    class(inner_product_class), target   :: newdot

    if (.not.associated(this%my_inner_product)) then
      this%my_inner_product => newdot
    else if (.not.associated(this%my_inner_product,target=newdot)) then
      call this%my_inner_product%D_times_g(this)
      call newdot%D_inverse_times_g(this)
      this%my_inner_product => newdot
    else
    endif
  end subroutine change_inner_product_to

   subroutine set_inner_product (this,newdot)
    implicit none
    class(gradient_class)                :: this
    class(inner_product_class), target   :: newdot
    this%my_inner_product => newdot
  end subroutine set_inner_product

  function inquire_inner_product (this)
    implicit none
    class(gradient_class)               :: this
    class(inner_product_class), pointer :: inquire_inner_product
    inquire_inner_product => this%my_inner_product
  end function inquire_inner_product
end module abstract_gradient
